// Copyright (c) 2010 Volvox Development Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Author: Konstantin Lepa <konstantin.lepa@gmail.com>

#ifndef XMLPARSER_XML_PARSER_HPP_
#define XMLPARSER_XML_PARSER_HPP_

/**
	@file xml-parser.hpp
	XML parser
*/

#include "xml-handler-base.hpp"

#include <expat.h>
#include <cstdio>
#include <memory>
#include <iconv.h>

#define C_XML_PARSER_BUFFER 8192

#ifndef XMLCALL
#define XMLCALL
#endif

namespace simplexml
{

/**
	@defgroup xmlgroup XML Parser
	@{
 */

/**
	XML Parser. Based on libexpat library.
 */
class XMLParser
{
public:
	/**
		@param handler - pointer to user handler object
		@param level - XML elements level
		@param includedFiles - list of included files,
		                       separated by '|' character
	 */
	XMLParser(simplexml::XMLHandlerBase *handler, UINT_32 level = 0,
	          const std::string &includedFiles = "");

	/**
		Initialise stream parser

		Call InitStream() before ParseStream() calls.
	 */
	void InitStream();

	/**
		Parse a block of the stream
		@param data - [in] data buffer
		@param len - [in] buffer size
	 */
	void ParseStream(CCHAR_P data, UINT_32 len);

	/**
		Finalise stream parser

		Don't forget to call CloseStream() after ParseStream() calls.
	 */
	void CloseStream();

	/**
		Parse document
		@param data - [in] data buffer
		@param len - [in] buffer size
	 */
	void ParseDocument(CCHAR_P data, UINT_32 len);

	/**
		Parse document from file
		@param fileName - [in] file stream
	 */
	void ParseFile(const std::string &fileName);

	~XMLParser();

private:
	bool inStream_;			///< Indicates stream parsing
	INT_32 level_;			///< Include tag level of recursion
	XML_Parser parser_;		///< Libexpat XML Parser object
	XMLHandlerBase *handler_;	///< User defined XML handler
	bool isIncluded_;		///< Indicates include usage
	std::string includedFiles_;	///< List of '|'-separated included files

	static UINT_16  KOI8_RMap[256];
	static UINT_16  CP866Map[256]; ///< Forward map for CP866 encoding
	static UINT_16  Windows_1251Map[256]; ///< Forward map for Windows-1251 encoding
	static UINT_16  CP1251Map[256]; ///< Forward map for CP1251 encoding
	static UINT_16  ISO_8859_5Map[256]; ///< Forward map for ISO-8859-5 encoding

	bool           iconvLoaded;
	iconv_t        iconvConverter;

	/**
		Start of element
		@param vThis - pointer to XML Parser object
		@param name - XML tag name
		@param attr - XML tag attributes
	 */
	static void XMLCALL StartElement(void *vThis, CCHAR_P name,
	                                 CCHAR_P *attr);

	/**
		End of Element
		@param vThis - pointer to XML Parser object
		@param name - XML tag name
	 */
	static void XMLCALL EndElement(void *vThis, CCHAR_P name);

	/**
		Charactes inside element
		@param vThis - pointer to XML Parser object
		@param data - Characters between tags
		@param len - Characters count
	 */
	static void XMLCALL Characters(void *vThis, CCHAR_P data, INT_32 len);

	/**
		Unknown encoding handler
	 */
	static int UnknownEncodingHandler(void *encHandlerData,
	                                  CCHAR_P name,
	                                  XML_Encoding *encInfo);
}; // class XMLParser

/* *@} */

} // namespace simplexml

#endif // XMLPARSER_XML_PARSER_HPP_
